# -*- coding: utf-8 -*-
import collections
import mimetypes
import itertools
import io
import random
import sys
import time
import urllib

from .http import Headers, Response
from .utils import to_text, to_bytes


def run_wsgi_app(app, environ):

    response = []
    buffer = []

    def start_response(status, headers, exc_info=None):
        if exc_info is not None:
            raise (exc_info[0], exc_info[1], exc_info[2])
        response[:] = [status, headers]
        return buffer.append

    app_rv = app(environ, start_response)
    app_iter = itertools.chain(buffer, app_rv)

    return app_iter, response[0], Headers.from_list(response[1])


def build_multipart_body(data, files):
    body = io.BytesIO()
    values = collections.OrderedDict()
    values.update(data or {})
    values.update(files or {})

    def write(content):
        body.write(to_bytes(content))
    boundary = '{0}{1}'.format(time.time(), random.random())

    for name, value in values.items():
        write('--{}\r\n'.format(boundary))
        write('Content-Disposition: form-data; name="{}"'.format(name))

        if isinstance(value, dict):  # file field
            filename = value.get('name', '')
            if filename:
                write('; filename="{}"'.format(filename))
            write('\r\n')
            content_type = (
                value.get('content_type') or
                mimetypes.guess_type(filename)[0] or
                'application/octet-stream')
            write('Content-Type: {}\r\n'.format(content_type))
            if not content_type.startswith('text'):
                write('Content-Transfer-Encoding: binary')
            value = value['file']

        write('\r\n\r\n')
        write(value)
        write('\r\n')
    write('--{}--\r\n'.format(boundary))
    length = body.tell()
    body.seek(0)
    return body, length, boundary


class Client(object):

    def __init__(self, app, host='localhost', port='80', cookies=None):
        self.app = app
        self.host = host
        self.environ_builder = EnvironBuilder()
        self.cookies = cookies or {}

    def open(self, path, method, params=None, data=None,
             files=None, headers=None, cookies=None,
             content_type='', charset='utf-8', follow_redirects=False):
        if isinstance(headers, dict):
            headers = Headers(headers)
        content_type = content_type or (headers or {}).get('Content-Type', '')
        cookies = cookies or {}
        cookies.update(self.cookies)
        body = None
        if files:
            body_reader, _, boundary = build_multipart_body(data, files)
            body = body_reader.read()
            body_reader.close()
            data = None
            content_type = 'multipart/form-data; boundary={}'.format(boundary)

        environ = self.environ_builder.build_environ(
            path=path, method=method, params=params,
            data=data, body=body, headers=headers,
            cookies=cookies,
            content_type=content_type, charset=charset
        )
        app_iter, status, headers = run_wsgi_app(self.app, environ)
        status_code = int(status[:3])
        response = Response(b''.join(app_iter), status_code=status_code,
                            headers=headers,
                            content_type=content_type)
        self.cookies.update({
            k: v.value
            for (k, v) in response.cookies.items()
        })
        if status_code in (301, 302, 303, 307) and follow_redirects:
            new_path = headers['Location']
            return self.open(new_path, 'GET')
        return response

    # get = functools.partialmethod(open, method='GET', data=None)
    def get(self, path, params=None, **kwargs):
        return self.open(path, method='GET', params=params, **kwargs)

    def options(self, path, **kwargs):
        return self.open(path, method='OPTIONS', **kwargs)

    def head(self, path, **kwargs):
        return self.open(path, method='HEAD', **kwargs)

    def trace(self, path, **kwargs):
        return self.open(path, method='TRACE', **kwargs)

    def post(self, path, **kwargs):
        return self.open(path, method='POST', **kwargs)

    def put(self, path, **kwargs):
        return self.open(path, method='PUT', **kwargs)

    def patch(self, path, **kwargs):
        return self.open(path, method='PATCH', **kwargs)

    def delete(self, path, **kwargs):
        return self.open(path, method='DELETE', **kwargs)


class EnvironBuilder(object):

    def __init__(self, host='localhost', port='80',
                 multithread=False, multiprocess=False,
                 run_once=False, environ_base=None,
                 url_scheme='http'):
        self.host = host
        self.port = str(port)
        if self.port == '80':
            self.http_host = self.host
        else:
            self.http_host = '{}:{}'.format(self.host, self.port)
        self.default_environ = {
            'SERVER_NAME': self.host,
            'GATEWAY_INTERFACE': 'CGI/1.1',
            'SERVER_PORT': self.port,
            'REMOTE_HOST': '',
            'CONTENT_LENGTH': '',
            'SCRIPT_NAME': '',
            'SERVER_PROTOCOL': 'HTTP/1.1',
            'REMOTE_ADDR': None,
            'REMOTE_PORT': None,

            'wsgi.version': (1, 0),
            'wsgi.url_scheme': url_scheme,
            'wsgi.errors': sys.stderr,
            'wsgi.multithread': multithread,
            'wsgi.multiprocess': multithread,
            'wsgi.run_once': run_once,
        }
        if environ_base:
            self.default_environ.update(environ_base)

    def build_environ(self, path='/', method='GET', params=None,
                      data=None, body=None, headers=None, cookies=None,
                      content_type='', charset='utf-8'):
        query_string = ''
        path = to_text(path)
        if '?' in path:
            path, query_string = path.split('?', 1)
        if isinstance(params, (dict, list, tuple)):
            query_string = query_string + '&' + urllib.parse.urlencode(params)

        if isinstance(data, dict):
            body = to_bytes(urllib.parse.urlencode(data), encoding=charset)
            content_type = 'application/x-www-form-urlencoded'
        elif data:
            body = to_bytes(data, encoding=charset)

        environ = self.default_environ.copy()
        environ.update({
            'REQUEST_METHOD': method,
            'PATH_INFO': path,
            'QUERY_STRING': query_string,
            'CONTENT_TYPE': content_type,
            'CONTENT_LENGTH': str(len(body or b'')),
            'wsgi.input': io.BytesIO(body),
        })
        # headers
        if headers is not None:
            for k, v in headers.to_dict().items():
                key = 'HTTP_' + k.replace('-', '_').upper()
                value = ', '.join(v)
                environ.update({key: value})
        environ.setdefault('HTTP_HOST', self.http_host)
        # cookies
        if cookies is not None:
            http_cookie = '; '.join('='.join([k, v])
                                    for (k, v) in cookies.items())
            environ.update({'HTTP_COOKIE': http_cookie})

        return environ



# -*- coding: utf-8 -*-
import mimetypes
import os

from .exceptions import NotFound
from .http import Response

http_methods = ('get', 'post', 'head', 'options',
                'delete', 'put', 'trace', 'patch')


class View:
    decorators = ()

    def dispatch_request(self, request, *args, **kwargs):
        method = request.method.lower()
        if method == 'head' and not hasattr(self, 'head'):
            method = 'get'
        view_func = getattr(self, method)
        return view_func(request, *args, **kwargs)

    @classmethod
    def as_view(cls, name=None, *class_args, **class_kwargs):

        def view(request, *args, **kwargs):
            instance = view.view_class(*class_args, **class_kwargs)
            return instance.dispatch_request(request, *args, **kwargs)

        for decorator in cls.decorators:
            view = decorator(view)

        view.view_class = cls
        view.__name__ = name or cls.__name__
        methods = []
        for method in http_methods:
            if hasattr(cls, method):
                methods.append(method.upper())
        methods = frozenset(methods)
        cls.methods = methods
        view.methods = methods
        return view


class StaticFilesView(View):

    def __init__(self, static_dir):
        self.static_dir = os.path.abspath(static_dir)

    def get(self, request, path):
        file_path = os.path.abspath(os.path.join(self.static_dir, path))
        if not file_path.startswith(self.static_dir):
            raise NotFound()
        if not os.path.isfile(file_path):
            raise NotFound()

        with open(file_path, 'rb') as fp:
            content = fp.read()
        content_type = mimetypes.guess_type(file_path)[0]
        content_type = content_type or 'application/octet-stream'
        return Response(content, content_type=content_type)



# -*- coding: utf-8 -*-
import abc
import collections
import contextlib
import logging

import psycopg2

logger = logging.getLogger(__name__)


class MetaData:
    tables = collections.OrderedDict()
    indexes = []

    @classmethod
    def index_sqls(cls):
        sqls = []
        for index in cls.indexes:
            sqls.append(index.to_sql())
        return ';\n'.join(sqls)

    @classmethod
    def create_all(cls, bind):
        connection = bind.connect()
        cursor = connection.cursor()
        for model in cls.tables.values():
            sql = model.table_sql()
            cursor.execute(sql)
        index_sql = cls.index_sqls()
        if index_sql:
            cursor.execute(index_sql)
        connection.commit()
        cursor.close()
        connection.close()

    @classmethod
    def drop_all(cls, bind):
        connection = bind.connect()
        cursor = connection.cursor()
        for table_name in cls.tables:
            cursor.execute('DROP TABLE IF EXISTS {}'.format(table_name))
        connection.commit()
        cursor.close()
        connection.close()


class Field(metaclass=abc.ABCMeta):
    def __init__(self, name=None, max_length=None, default=None,
                 server_default=None, unique=False, nullable=True,
                 index=False, primary_key=False, foreign_key=None):
        self.name = name
        self.max_length = max_length
        self.default = default
        if server_default is True or server_default is False:
            server_default = str(server_default).upper()
        self.server_default = server_default
        self.unique = unique
        self.nullable = nullable
        self.index = index
        self.primary_key = primary_key
        self.foreign_key = foreign_key

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if self._name not in instance.__dict__:
            value = getattr(owner, self._name)
            return value.default_value()
        else:
            return instance.__dict__[self._name]

    def __set__(self, instance, value):
        instance.__dict__[self._name] = value

    def __lt__(self, value):
        return '{.sql_column} < %s'.format(self), value

    def __le__(self, value):
        return '{.sql_column} <= %s'.format(self), value

    def __eq__(self, value):
        return '{.sql_column} = %s'.format(self), value

    def __ne__(self, value):
        return '{.sql_column} != %s'.format(self), value

    def __gt__(self, value):
        return '{.sql_column} > %s'.format(self), value

    def __ge__(self, value):
        return '{.sql_column} >= %s'.format(self), value

    def like(self, value):
        return '{.sql_column} LIKE %s'.format(self), value

    @property
    def desc(self):
        return '{.sql_column} DESC'.format(self)

    def is_(self, value):
        return '{.sql_column} IS %s'.format(self), value

    def is_not(self, value):
        return '{.sql_column} IS NOT %s'.format(self), value

    def not_in(self, value):
        '''
        :type value: tuple
        '''
        return '{.sql_column} NOT IN %s'.format(self), value

    def in_(self, value):
        '''
        :type value: tuple
        '''
        return '{.sql_column} IN %s'.format(self), value

    def to_sql(self):
        sql = '{0.name} {0.data_type}'.format(self)
        if self.max_length is not None:
            sql += '({0.max_length})'.format(self)
        if self.server_default:
            sql += ' DEFAULT {0.server_default}'.format(self)
        if not self.nullable:
            sql += ' NOT NULL'
        if self.unique:
            sql += ' UNIQUE'
        if self.primary_key:
            sql += ' PRIMARY KEY'
        if self.foreign_key is not None:
            sql += ' {}'.format(self.foreign_key.to_sql())
        return sql

    def default_value(self):
        value = self.default
        if callable(self.default):
            value = self.default()
        if self.default is None:
            if isinstance(self, TextField):
                value = 'None'
        return value

    def name_sql(self):
        return '{}.{}'.format(self.model.table_name, self.name)


def _collect_fields(attr_dict, model):
    fields = []
    for field in attr_dict.values():
        if isinstance(field, Field):
            fields.append(field)
            field.model = model
            field.table_name = model.table_name
            field.sql_column = '{}.{}'.format(field.table_name, field.name)
    return fields


def _get_table_name(attr_dict):
    return attr_dict.get('__tablename__')


def _auto_column_name(attr_dict):
    for attr, attr_value in attr_dict.items():
        if isinstance(attr_value, Field):
            attr_value._name = attr
            if attr_value.name is None:
                attr_value.name = attr


def _collect_indexes(table_name, attr_dict):
    indexes = []
    for attr, attr_value in attr_dict.items():
        if isinstance(attr_value, Field):
            if not any([attr_value.unique, attr_value.primary_key,
                        attr_value.foreign_key]) and attr_value.index:
                indexes.append(attr_value)
    for field in indexes:
        column_name = field.name
        name = 'index_{}_{}'.format(table_name, column_name)
        index = Index(name, table_name, column_name, unique=field.unique)
        MetaData.indexes.append(index)


class ModelMetaClass(type):

    def __init__(cls, name, bases, attr_dict):
        table_name = _get_table_name(attr_dict)
        if table_name:
            cls.table_name = table_name
            MetaData.tables[table_name] = cls
            _auto_column_name(attr_dict)
            cls.fields = _collect_fields(attr_dict, cls)
            for field in cls.fields:
                if field.primary_key:
                    cls.pk_name = field.name
                    break
            _collect_indexes(table_name, attr_dict)

    @classmethod
    def __prepare__(cls, name, bases):
        """让 attr_dict 有序"""
        return collections.OrderedDict()


class Model(metaclass=ModelMetaClass):

    def __init__(self, **kwargs):
        self.__dict__.update(self.default_dict())
        for kwg, value in kwargs.items():
            setattr(self, kwg, value)
        for field in self.fields:
            if field.primary_key:
                self.pk_field = field
                break

    def default_dict(self):
        return {field._name: field.default_value() for field in self.fields}

    @classmethod
    def table_sql(cls):
        column_sqls = ',\n    '.join(field.to_sql() for field in cls.fields)
        sql = '''
CREATE TABLE {table_name} (
    {column_sqls}
);
'''.format(table_name=cls.table_name, column_sqls=column_sqls)
        return sql

    def sql_values(self):
        values_dict = collections.OrderedDict()
        for field in self.fields:
            if field is self.pk_field:
                continue
            value = getattr(self, field._name, field.default_value())
            if value is not None:
                values_dict[field.name] = value
        return values_dict


class TextField(Field):
    data_type = 'text'


class CharField(TextField):
    data_type = 'varchar'

    def __init__(self, max_length=None, **kwargs):
        super(CharField, self).__init__(max_length=max_length, **kwargs)


class IntegerField(Field):
    data_type = 'integer'


class DateField(Field):
    data_type = 'date'


class DateTimeField(Field):
    data_type = 'timestamp'


class BooleanField(Field):
    data_type = 'boolean'


class UUIDField(Field):
    data_type = 'uuid'


class JSONField(Field):
    data_type = 'json'


class AutoField(Field):
    data_type = 'serial'


class ForeignKey:
    """

    :param column: A single target column for the key relationship.
                   A column name as a string: `table_name.column_name`
    :param onupdate: Optional string. If set, emit ON UPDATE <value>
                     when issuing DDL for this constraint. Typical values
                     include CASCADE, DELETE and RESTRICT.
    :param ondelete: Optional string. If set, emit ON DELETE <value>
                     when issuing DDL for this constraint. Typical values
                     include CASCADE, DELETE and RESTRICT.
    """
    def __init__(self, column, onupdate=None, ondelete=None):
        self.table_name, self.column_name = column.split('.')
        self.onupdate = onupdate
        self.ondelete = ondelete

    def to_sql(self):
        sql = 'REFERENCES {0.table_name} ({0.column_name})'.format(self)
        if self.onupdate:
            sql += ' ON UPDATE {0.onupdate}'.format(self)
        if self.ondelete:
            sql += ' ON DELETE {0.ondelete}'.format(self)
        return sql


class Index:
    def __init__(self, name, table_name, column_name, unique=False):
        self.name = name
        self.table_name = table_name
        self.column_name = column_name
        self.unique = unique

    def to_sql(self):
        sql = 'CREATE'
        if self.unique:
            sql += ' UNIQUE'
        sql += (' INDEX {0.name} ON {0.table_name} ({0.column_name})'
                ).format(self)
        return sql + ';'


class Engine:

    def __init__(self, uri):
        self.uri = uri

    def connect(self):
        self.connection = psycopg2.connect(self.uri)
        return self.connection

    def close(self):
        self.connection.close()


class Session:
    _bind = None

    def __init__(self, bind=None):
        self.bind = bind or self._bind
        if self.bind is not None:
            self.connect()

    @classmethod
    def configure(cls, bind):
        cls._bind = bind

    def connect(self):
        if (not hasattr(self, 'connection')) or self.connection.closed:
            self.connection = self.bind.connect()
        if (not hasattr(self, 'cursor')) or self.cursor.closed:
            self.cursor = self.connection.cursor()

    def execute(self, sql, args):
        logger.debug('execute sql: %s, args: %s', sql, args)
        self.connect()
        return self.cursor.execute(sql, args)

    def fetchone(self):
        return self.cursor.fetchone()

    def fetchmany(self, size=None):
        if size is None:
            return self.cursor.fetchmany()
        else:
            return self.cursor.fetchmany(size)

    def fetchall(self):
        return self.cursor.fetchall()

    def commit(self):
        self.connection.commit()

    def rollback(self):
        self.connection.rollback()

    def close(self):
        self.cursor.close()
        self.connection.close()

    def insert(self, instance):
        pk_name = instance.pk_field.name
        sql_values = instance.sql_values()
        columns = ', '.join('{0}'.format(k) for k in sql_values)
        values = ', '.join('%s' for k in sql_values)
        sql = (
            'INSERT INTO {table_name} ({columns}) VALUES ({values}) '
            'RETURNING {pk_name};').format(
            table_name=instance.table_name, columns=columns,
            values=values, pk_name=pk_name
        )
        args = list(sql_values.values())
        self.execute(sql, args)
        pk_value = self.fetchone()[0]
        setattr(instance, instance.pk_field._name, pk_value)

    def update(self, instance):
        pk_name = instance.pk_field.name
        pk_value = getattr(instance, pk_name)
        sql_values = instance.sql_values()
        columns = ', '.join('{0} = %s'.format(k) for k in sql_values)
        sql = 'UPDATE {table_name} SET {columns} WHERE {pk_name} = %s;'.format(
            table_name=instance.table_name, pk_name=pk_name,
            columns=columns
        )
        args = list(sql_values.values()) + [pk_value]
        self.execute(sql, args)

    def delete(self, instance):
        pk_name = instance.pk_field.name
        pk_value = getattr(instance, pk_name)
        sql = 'DELETE FROM {table_name} WHERE {pk_name} = %s'.format(
            table_name=instance.table_name, pk_name=pk_name
        )
        args = (pk_value,)
        self.execute(sql, args)
        setattr(instance, pk_name, None)

    def query(self, model):
        return QuerySet(self, model)

    @contextlib.contextmanager
    def transaction(self):
        try:
            yield
        except:
            self.rollback()
            raise
        else:
            self.commit()


class QuerySet:

    def __init__(self, session, model):
        self.session = session
        self.model = model
        self.wheres = []
        self._limit = None
        self._offset = None
        self._order_by = []

    def clone(self):
        queryset = type(self)(self.session, self.model)
        queryset.wheres.extend(self.wheres)
        queryset._limit = self._limit
        queryset._offset = self._offset
        queryset._order_by = self._order_by
        return queryset

    def limit(self, number):
        self._limit = number
        return self.clone()

    def offset(self, number):
        self._offset = number
        return self.clone()

    def order_by(self, *args):
        orders = []
        for arg in args:
            if isinstance(arg, Field):
                exp = '{0}.{1}'.format(self.model.table_name, arg.name)
            else:
                exp = str(arg)
            orders.append(exp)
        self._order_by.extend(orders)
        return self.clone()

    def filter(self, *args, **kwargs):
        wheres = []
        for arg in args:
            wheres.append(arg)
        for kw, value in kwargs.items():
            wheres.append(('{} = %s'.format(kw), value))
        self.wheres.extend(wheres)
        return self.clone()

    def count(self):
        sql, args = self._build_select_sql(count=True)
        self.session.execute(sql, args)
        return self.session.fetchone()[0]

    def update(self, **kwargs):
        sql_values = collections.OrderedDict()
        for kw, value in kwargs.items():
            for field in self.model.fields:
                if field._name == kw:
                    sql_values[field.name] = value
                    break
            else:
                raise KeyError('{} is invalid field name'.format(kw))
        sql, args = self._build_update_sql(sql_values)
        self.session.execute(sql, args)
        self._clean_data()

    def delete(self):
        sql, args = self._build_delete_sql()
        self.session.execute(sql, args)
        self._clean_data()

    def _build_where_sql(self):
        where = ' AND '.join(x[0] for x in self.wheres)
        args = [x[1] for x in self.wheres if len(x) > 1]
        if where:
            where = 'WHERE ' + where
        return where, args

    def _build_limit_sql(self):
        if self._limit is not None:
            return 'LIMIT {}'.format(self._limit)
        return ''

    def _build_offset_sql(self):
        if self._offset is not None:
            return 'OFFSET {}'.format(self._offset)
        return ''

    def _build_order_by_sql(self):
        if self._order_by:
            order_by = ', '.join(self._order_by)
            return 'ORDER BY {0}'.format(order_by)
        return ''

    def _build_select_sql(self, count=False):
        table_name = self.model.table_name
        where, args = self._build_where_sql()
        if count:
            column_names = 'COUNT(*)'
            order_by = ''
            offset = ''
            limit = ''
        else:
            column_names = ', '.join(
                '{column_name} AS {table_name}_{column_name}'.format(
                    column_name=field.name, table_name=table_name
                )
                for field in self.model.fields
            )
            order_by = self._build_order_by_sql()
            offset = self._build_offset_sql()
            limit = self._build_limit_sql()

        sql = (
            'SELECT {column_names} FROM {table_name} {where}'
            ' {offset} {limit} {order_by};'.format(
                column_names=column_names, table_name=table_name, where=where,
                offset=offset, limit=limit, order_by=order_by
            )
        )
        return sql, args

    def _build_update_sql(self, sql_values):
        table_name = self.model.table_name
        where, args = self._build_where_sql()
        columns = ', '.join('{0} = %s'.format(k) for k in sql_values)
        sql = (
            'UPDATE {table_name} SET {columns} {where};'.format(
                table_name=table_name, columns=columns, where=where,
            )
        )
        args = list(sql_values.values()) + args
        return sql, args

    def _build_delete_sql(self):
        table_name = self.model.table_name
        where, args = self._build_where_sql()
        sql = 'DELETE FROM {table_name} {where};'.format(
            table_name=table_name,  where=where
        )
        return sql, args

    def _execute(self):
        if hasattr(self, '_data'):
            return
        sql, args = self._build_select_sql()
        self.session.execute(sql, args)
        self._data = []
        for row in self.session.fetchall():
            instance = self.model()
            for nu, value in enumerate(row):
                setattr(instance, self.model.fields[nu].name, value)
            self._data.append(instance)

    def _clean_data(self):
        if hasattr(self, '_data'):
            delattr(self, '_data')

    def __len__(self):
        if self._limit is not None or self._offset is not None:
            self._execute()
            return len(self._data)
        return self.count()

    def __getitem__(self, index):
        self._execute()
        return self._data[index]

    def __iter__(self):
        self._execute()
        return iter(self._data)



# -*- coding: utf-8 -*-
import abc
import wsgiref.simple_server

from . import wsgi_server


class ServerAdapter(metaclass=abc.ABCMeta):
    def __init__(self, host='127.0.0.1', port=5000, **options):
        self.host = host
        self.port = port
        self.options = options

    @abc.abstractmethod
    def run(self, app):
        pass


class BustardServer(ServerAdapter):

    def run(self, app):
        httpd = wsgi_server.make_server(
            (self.host, self.port), app, **self.options
        )
        httpd.serve_forever()


class WSGIRefServer(ServerAdapter):

    def run(self, app):
        httpd = wsgiref.simple_server.make_server(
            self.host, self.port, app, **self.options
        )
        httpd.serve_forever()


class WerkzeugfServer(ServerAdapter):

    def run(self, app):
        from werkzeug.serving import run_simple
        run_simple(self.host, self.port, app, self.options)



# -*- coding: utf-8 -*-
import abc
import collections
import uuid


class SessionBase(collections.UserDict):

    def __init__(self, session_key=None):
        super(SessionBase, self).__init__()
        self.session_key = session_key
        self.modified = False

    def is_empty(self):
        return not (self.session_key and self.data)

    def __setitem__(self, key, value):
        self.data[key] = value
        self.modified = True

    def __delitem__(self, key):
        del self.data[key]
        self.modified = True

    def pop(self, key, default=None):
        self.modified = self.modified or key in self.data
        return self.data.pop(key, default)

    def setdefault(self, key, value):
        if key in self.data:
            return self.data[key]
        else:
            self.modified = True
            self.data[key] = value
            return value

    def update(self, dict_):
        self.data.update(dict_)
        self.modified = True

    def clear(self):
        self.data = {}
        self.accessed = True
        self.modified = True

    def _get_or_create_session_key(self):
        if not self.session_key or not self.exists(self.session_key):
            self.session_key = self._new_session_key()
        return self.session_key

    def _new_session_key(self):
        return uuid.uuid4().hex

    @abc.abstractmethod
    def create(self):
        pass

    @abc.abstractmethod
    def exists(self, session_key=None):
        pass

    @abc.abstractmethod
    def save(self):
        pass

    @abc.abstractmethod
    def delete(self, session_key=None):
        pass


def before_request_hook(request, view_func, app):
    session_key = request.cookies.get(app.config['SESSION_COOKIE_NAME'])
    request.session = app.session_class(session_key=session_key)


def after_request_hook(request, response, view_func, app):
    session = request.session
    if not session.modified:
        return

    config = app.config
    cookie_name = config['SESSION_COOKIE_NAME']
    if session.is_empty():
        if cookie_name in request.cookies:
            session.delete()
            response.delete_cookie(cookie_name)
    else:
        session.save()
        response.set_cookie(
            cookie_name, value=session.session_key,
            max_age=config['SESSION_COOKIE_MAX_AGE'],
            expires=None, path=config['SESSION_COOKIE_PATH'],
            domain=config['SESSION_COOKIE_DOMAIN'],
            secure=config['SESSION_COOKIE_SECURE'],
            httponly=config['SESSION_COOKIE_HTTPONLY']
        )
    return response


class MemorySession(SessionBase):
    _sessions = {}

    def __init__(self, session_key=None):
        super(MemorySession, self).__init__(session_key=session_key)
        self.load(session_key)

    def load(self, session_key):
        _sessions = self.__class__._sessions
        if session_key not in _sessions:
            self.create()
        else:
            self.data = _sessions[self.session_key]

    def exists(self, key):
        _sessions = self.__class__._sessions
        return key in _sessions

    def create(self):
        self.modified = True
        self.save(must_create=True)

    def save(self, must_create=False):
        session_key = self._get_or_create_session_key()
        _sessions = self.__class__._sessions
        if self.data or must_create:
            if session_key in _sessions and not self.modified:
                self.data = _sessions[self.session_key]
            else:
                _sessions[self.session_key] = self.data
        else:
            self.delete()

    def delete(self, session_key=None):
        _sessions = self.__class__._sessions
        session_key = session_key or self.session_key
        _sessions.pop(session_key, None)
        self.modified = True



# -*- coding: utf-8 -*-
import binascii
import base64
import collections
import urllib


class MultiDict(collections.UserDict):

    def getlist(self, key):
        return self.data[key]

    def to_dict(self):
        return self.data

    def __getitem__(self, key):
        return self.data[key][-1]

    def __setitem__(self, key, value):
        if isinstance(value, (list, tuple)):
            self.data[key] = list(value)
        else:
            self.data[key] = [value]

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.data)


def json_dumps_default(obj):
    if isinstance(obj, collections.UserDict):
        return obj.to_dict()
    return obj


def parse_query_string(query_string, encoding='utf-8'):
    query_dict = collections.defaultdict(list)
    for query_item in query_string.split('&'):
        if '=' not in query_item:
            continue
        keyword, value = query_item.split('=', 1)
        value = urllib.parse.unquote_plus(value)
        query_dict[keyword].append(to_text(value, encoding=encoding))
    return query_dict


def to_header_key(key):
    return '-'.join(x.capitalize() for x in key.split('-'))


def to_text(st, encoding='utf-8'):
    if isinstance(st, str):
        return st
    elif isinstance(st, collections.ByteString):
        return st.decode(encoding)
    else:
        return str(st)


def to_bytes(bt, encoding='utf-8'):
    if isinstance(bt, collections.ByteString):
        return bt
    elif isinstance(bt, str):
        return bt.encode(encoding)
    else:
        return bytes(bt)


class Authorization:

    def __init__(self, _type, username, password):
        self.type = _type
        self.username = username
        self.password = password

    def __eq__(self, other):
        return (
            self.type == other.type and
            self.username == other.username and
            self.password == other.password
        )

    __hash__ = object.__hash__

    def __repr__(self):
        return '{}(type:{}, username:{})'.format(
            self.__class__.__name__, self.type, self.username
        )


def parse_basic_auth_header(value):
    try:
        auth_type, auth_info = to_bytes(value).split(None, 1)
    except ValueError:
        return
    auth_type = auth_type.lower()

    if auth_type == b'basic':
        try:
            username, password = base64.b64decode(auth_info).split(b':', 1)
        except (binascii.Error, ValueError):
            return

        return Authorization(
            to_text(auth_type),
            username=to_text(username),
            password=to_text(password)
        )


def cookie_date():
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import datetime
import io
import socket
import sys
import time
import urllib

from .utils import to_text, to_bytes


class WSGIServer(object):
    address_family = socket.AF_INET
    socket_type = socket.SOCK_STREAM
    request_queue_size = 5
    allow_reuse_address = True
    default_request_version = 'HTTP/1.1'
    server_version = 'WSGIServer/0.1'
    weekdayname = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
    monthname = [None,
                 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']

    def __init__(self, server_address):
        # 创建 socket
        self.socket = socket.socket(self.address_family, self.socket_type)
        # 绑定
        self.server_bind(server_address)
        # 监听
        self.server_activate()
        # 基本的 environ
        self.setup_environ()
        self.headers_set = []

    def server_bind(self, server_address):
        if self.allow_reuse_address:
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        self.socket.bind(server_address)
        self.server_address = self.socket.getsockname()

        host, port = self.socket.getsockname()[:2]
        self.server_name = socket.getfqdn(host)
        self.server_port = port

    def server_activate(self):
        self.socket.listen(self.request_queue_size)

    def setup_environ(self):
        """https://www.python.org/dev/peps/pep-0333/#environ-variables"""
        # Set up base environment
        env = self.base_environ = {}
        env['SERVER_NAME'] = self.server_name
        env['GATEWAY_INTERFACE'] = 'CGI/1.1'
        env['SERVER_PORT'] = str(self.server_port)
        env['REMOTE_HOST'] = ''
        env['CONTENT_LENGTH'] = ''
        env['SCRIPT_NAME'] = ''

    def get_app(self):
        return self.application

    def set_app(self, application):
        self.application = application

    def serve_forever(self):
        while 1:
            self.handle_one_request()
            self.client_connection.close()

    def handle_one_request(self):
        self.client_connection, self.client_address = self.socket.accept()
        # self.raw_request = raw_request = self.client_connection.recv(65537)
        init_read = 65536
        self.raw_request = raw_request = self.client_connection.recv(init_read)

        self.parse_request(raw_request)
        self.parse_headers(raw_request)
        length = int(self.headers.get('Content-Length', '0'))
        while len(self.raw_request) == length:
            self.raw_request += self.client_connection.recv(init_read)

        env = self.get_environ()

        print('[%s] "%s %s %s"' % (
            datetime.datetime.now(), env['REQUEST_METHOD'],
            env['PATH_INFO'], env['SERVER_PROTOCOL'],
        ))

        result = self.application(env, self.start_response)
        self.finish_response(result)

    def parse_request(self, raw_request):
        # GET /foo?a=1&b=2 HTTP/1.1
        first_line = to_text(raw_request.split(b'\r\n', 1)[0].strip())
        (self.request_method,   # GET
         self.path,             # /foo?a=1&b=2
         self.request_version   # HTTP/1.1
         ) = first_line.split()

    def parse_headers(self, raw_request):
        header_string = to_text(raw_request.split(b'\r\n\r\n', 1)[0])
        self.headers = headers = {}
        for header in header_string.splitlines()[1:]:
            k, v = header.split(':', 1)
            if headers.get(k):
                headers[k] += ', ' + v.strip()  # 多个同名 header
            else:
                headers[k] = v.strip()

    def get_environ(self):
        """https://www.python.org/dev/peps/pep-0333/#environ-variables"""
        env = self.base_environ.copy()
        env['REQUEST_METHOD'] = self.request_method

        if '?' in self.path:
            path, query = self.path.split('?', 1)
        else:
            path, query = self.path, ''
        env['PATH_INFO'] = urllib.parse.unquote(path)
        env['QUERY_STRING'] = query

        env['CONTENT_TYPE'] = self.headers.get('Content-Type', '')
        env['CONTENT_LENGTH'] = self.headers.get('Content-Length', '0')

        env['SERVER_PROTOCOL'] = self.request_version
        env['REMOTE_ADDR'] = self.client_address[0]
        env['REMOTE_PORT'] = self.client_address[1]

        env['wsgi.version'] = (1, 0)
        env['wsgi.url_scheme'] = 'http'
        env['wsgi.input'] = io.BytesIO(self.raw_request)
        env['wsgi.errors'] = sys.stderr
        env['wsgi.multithread'] = False
        env['wsgi.multiprocess'] = True
        env['wsgi.run_once'] = False

        for k, v in self.headers.items():
            k = k.replace('-', '_').upper()
            if k in env:
                continue
            env['HTTP_' + k] = v
        return env

    def start_response(self, status, headers, exc_info=None):
        server_headers = [
            ('Date', self.date_time_string()),
            ('Server', self.version_string()),
        ]
        headers = list(headers) + server_headers

        if exc_info:
            try:
                if self.headers_set:
                    # Re-raise original exception if headers sent
                    raise (exc_info[0], exc_info[1], exc_info[2])
            finally:
                exc_info = None     # avoid dangling circular ref

        self.headers_set[:] = [status, headers]

    def finish_response(self, body):
        try:
            status, headers = self.headers_set
            # status line
            response = (
                to_bytes(self.default_request_version) +
                b' ' +
                to_bytes(status) +
                b'\r\n'
            )
            # headers
            response += b'\r\n'.join([to_bytes(': '.join(x)) for x in headers])
            response += b'\r\n\r\n'
            # body
            for d in body:
                response += d
            self.client_connection.sendall(response)
        finally:
            self.client_connection.close()

    def version_string(self):
        return self.server_version

    def date_time_string(self, timestamp=None):
        if timestamp is None:
            timestamp = time.time()
        year, month, day, hh, mm, ss, wd, y, z = time.gmtime(timestamp)
        s = '%s, %02d %3s %4d %02d:%02d:%02d GMT' % (
            self.weekdayname[wd],
            day, self.monthname[month], year,
            hh, mm, ss
        )
        return s


def make_server(server_address, application):
    server = WSGIServer(server_address)
    server.set_app(application)
    return server


if __name__ == '__main__':
    if len(sys.argv) < 2:
        sys.exit('Provide a WSGI application object as module:callable')
    app_path = sys.argv[1]
    module, application = app_path.split(':')
    module = __import__(module)
    application = getattr(module, application)

    SERVER_ADDRESS = (HOST, PORT) = '', 8888
    httpd = make_server(SERVER_ADDRESS, application)
    print('WSGIServer: Serving HTTP on port {port} ...\n'.format(port=PORT))

    httpd.serve_forever()



# -*- coding: utf-8 -*-
from .constants import NOTFOUND_HTML
from .http import Response


class HTTPException(Exception):
    def __init__(self, response):
        self.response = response


class NotFound(HTTPException):
    def __init__(self):
        self.response = Response(NOTFOUND_HTML, status_code=404)



# -*- coding: utf-8 -*-
import cgi
from http.cookies import SimpleCookie
import io
import json

from .constants import HTTP_STATUS_CODES
from .utils import (
    json_dumps_default, MultiDict, parse_query_string,
    to_header_key, to_text, to_bytes, parse_basic_auth_header
)


class Request(object):

    def __init__(self, environ):
        self.environ = environ

    @property
    def method(self):
        """``GET``, ``POST`` etc."""
        return self.environ['REQUEST_METHOD']

    @property
    def path(self):
        return self.environ['PATH_INFO']

    @property
    def host(self):
        return self.environ['HTTP_HOST']

    @property
    def scheme(self):
        return self.environ['wsgi.url_scheme']

    @property
    def url(self):
        return '{scheme}://{host}{path}'.format(
            scheme=self.scheme, host=self.host, path=self.path
        )

    @property
    def remote_addr(self):
        return self.environ.get('REMOTE_ADDR', '')

    @property
    def form(self):
        if self.method not in ['POST', 'PUT', 'PATCH', 'DELETE']:
            return {}
        content_type = self.content_type
        if (
            content_type.startswith('multipart/form-data; boundary=') or
            content_type.startswith('application/x-www-form-urlencoded')
        ):
            self.parse_form_data()
            return MultiDict(self._form)
        else:
            return {}

    def parse_form_data(self):
        if hasattr(self, '_form'):
            return

        fields = cgi.FieldStorage(fp=self.stream, environ=self.environ)
        _form = {}
        _files = {}
        if fields.length > 0 and fields.list:
            for key in fields:
                values = fields[key]
                if isinstance(values, list):
                    _form[key] = [x.value for x in values]
                elif 'Content-Type' in values.headers:
                    _files[key] = File(values.value, values.filename,
                                       values.type)
                else:
                    _form[key] = [values.value]
        self._form = _form
        self._files = _files

    @property
    def args(self):
        query_string = self.environ['QUERY_STRING']
        return MultiDict(parse_query_string(query_string))

    @property
    def values(self):
        raise NotImplementedError

    @property
    def cookies(self):
        """Request cookies

        :rtype: dict
        """
        http_cookie = self.environ.get('HTTP_COOKIE', '')
        _cookies = {
            k: v.value
            for (k, v) in SimpleCookie(http_cookie).items()
        }
        return _cookies

    @property
    def headers(self):
        _headers = {
            to_header_key(key.replace('HTTP_', '', 1).replace('_', '-')): value
            for key, value in self.environ.items()
            if key.startswith('HTTP_')
        }
        _headers.setdefault('Content-Type', self.content_type)
        _headers.setdefault('Content-Length', self.content_length)
        return Headers(_headers)

    @property
    def data(self, as_text=False, encoding='utf-8'):
        if hasattr(self, '_content'):
            return self._content

        if self.content_type in [
            'application/x-www-form-urlencoded',
            'multipart/form-data',
        ]:
            content = b''
        else:
            content = self.stream.read(int(self.content_length or 0))
        self._content = content
        if as_text:
            content = content.decode(encoding)
        return content

    @property
    def files(self):
        if self.method not in ['POST', 'PUT']:
            return {}
        self.parse_form_data()
        return self._files

    @property
    def stream(self):
        return self.environ['wsgi.input']

    def get_json(self, encoding='utf-8'):
        content = self.data.decode(encoding)
        try:
            return json.loads(content)
        except ValueError:
            return

    @property
    def content_type(self):
        return self.environ.get('CONTENT_TYPE', '')

    @property
    def content_length(self):
        return self.environ.get('CONTENT_LENGTH', '')

    @property
    def is_json(self):
        content_type = self.content_type
        if content_type.startswith('application/json') or (
                content_type.startswith('application/') and
                content_type.endswith('+json')
                ):
            return True
        return False

    @property
    def is_ajax(self):
        """The ``X-Requested-With`` header equal to ``HttpRequest`` """
        requested_with = self.headers.get('HTTP_X_REQUESTED_WITH', '').lower()
        return requested_with == 'xmlhttprequest'

    @property
    def authorization(self):
        headers = self.headers
        auth_header_value = headers.get('Authorization', '')
        if len(auth_header_value.split()) > 1:
            return parse_basic_auth_header(auth_header_value)


class Response(object):

    def __init__(self, content=b'', status_code=200,
                 content_type='text/html; charset=utf-8',
                 headers=None):
        self._content = content
        self._status_code = status_code
        _headers = headers or {}
        _headers.setdefault('Content-Type', content_type)
        if isinstance(_headers, Headers):
            self._headers = _headers
        else:
            self._headers = Headers(_headers)
        self._cookies = SimpleCookie()
        self._load_cookies_from_headers()

    def _load_cookies_from_headers(self):
        cookies = self._headers.to_dict().pop('Set-Cookie', [])
        for cookie in cookies:
            self._cookies.load(cookie)

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, value):
        if isinstance(value, str):
            value = value.encode('utf-8')
        self._content = value
    body = data = content

    def get_data(self):
        return self._content

    @property
    def content_type(self, value):
        return self.headers.get('Content-Type', '')

    @content_type.setter
    def content_type(self, value):
        self.headers['Content-Type'] = value

    @property
    def content_length(self):
        return int(self.headers.get('Content-Length', '0'))

    @property
    def status_code(self):
        return self._status_code

    @status_code.setter
    def status_code(self, value):
        self._status_code = value

    @property
    def status(self):
        code = self._status_code
        return response_status_string(code)

    @property
    def headers(self):
        return self._headers

    @headers.setter
    def headers(self, value):
        self._headers = Headers(value)

    @property
    def content_type(self):
        return self._headers.get('Content-Type', '')

    @content_type.setter
    def content_type(self, value):
        self._headers['Content-Type'] = value

    @property
    def cookies(self):
        return self._cookies

    def set_cookie(self, key, value='', max_age=None, expires=None, path='/',
                   domain=None, secure=False, httponly=False):
        cookie = cookie_dump(
            key, value=value, max_age=max_age, expires=expires, path=path,
            domain=domain, secure=secure, httponly=httponly
        )
        self._cookies.load(cookie)

    def delete_cookie(self, key, max_age=0,
                      expires='Thu, 01-Jan-1970 00:00:00 GMT'):
        self.set_cookie(key, value='', max_age=max_age, expires=expires)

    @property
    def headers_list(self):
        # normal headers
        headers_list = list(self.headers.to_list())

        # set-cookies
        headers_list.extend(
            ('Set-Cookie', value.OutputString())
            for value in self.cookies.values()
        )
        return headers_list

    def json(self):
        return json.loads(to_text(self.data))

    def __repr__(self):
        return '<{} [{}]>'.format(self.__class__.__name__, self.status_code)


def cookie_dump(key, value='', max_age=None, expires=None, path='/',
                domain=None, secure=False, httponly=False):
    """
    :rtype: ``Cookie.SimpleCookie``
    """
    cookie = SimpleCookie()
    cookie[key] = value
    for attr in ('max_age', 'expires', 'path', 'domain',
                 'secure', 'httponly'):
        attr_key = attr.replace('_', '-')
        attr_value = locals()[attr]
        if attr_value:
            cookie[key][attr_key] = attr_value
    return cookie


def response_status_string(code):
    """e.g. ``200 OK`` """
    mean = HTTP_STATUS_CODES.get(code, 'unknown').upper()
    return '{code} {mean}'.format(code=code, mean=mean)


def jsonify(*args, **kwargs):
    data = json.dumps(dict(*args, **kwargs), indent=2, sort_keys=True,
                      separators=(', ', ': '), default=json_dumps_default)
    data = data.encode('utf-8')
    response = Response(data + b'\n', content_type='application/json')
    response.headers['Content-Length'] = str(len(response.data))
    return response


def redirect(url, code=302):
    response = Response(status_code=code)
    response.headers['Location'] = url
    return response


class Headers(MultiDict):

    def add(self, key, value):
        key = to_text(to_header_key(key))
        if isinstance(value, (tuple, list)):
            self.data.setdefault(key, []).extend(map(to_text, value))
        else:
            self.data.setdefault(key, []).append(to_text(value))

    def set(self, key, value):
        self.__setitem__(key, value)

    def get_all(self, key):
        key = to_header_key(key)
        return self.data[key]

    @classmethod
    def from_list(cls, headers_list):
        headers = cls()
        for (k, v) in headers_list:
            headers.add(k, v)
        return headers

    def to_list(self):
        return [
            (k, v)
            for k, values in self.to_dict().items()
            for v in values
        ]

    def __getitem__(self, key):
        key = to_header_key(key)
        return super(Headers, self).__getitem__(key)

    def __setitem__(self, key, value):
        key = to_text(to_header_key(key))
        if isinstance(value, (list, tuple)):
            value = list(map(to_text, value))
        else:
            value = to_text(value)
        super(Headers, self).__setitem__(key, value)


class File:

    def __init__(self, data, filename,
                 content_type='application/octet-stream'):
        self.file = io.BytesIO(to_bytes(data))
        self.file.name = filename
        self.file.content_type = content_type

    def __getattr__(self, attr):
        return getattr(self.file, attr)



# -*- coding: utf-8 -*-
import collections
import inspect
import os

from .constants import CONFIGURE
from .exceptions import HTTPException, NotFound
from .http import Request, Response
from .router import Router
from .template import Template
from .testing import Client
from .utils import to_bytes
from .servers import WSGIRefServer
from . import sessions


class Bustard(object):
    session_class = sessions.MemorySession
    before_request_hooks = (sessions.before_request_hook,)
    after_request_hooks = (sessions.after_request_hook,)

    def __init__(self, name='', template_dir='',
                 template_default_context=None):
        self.name = name
        self._router = Router()
        self.template_dir = template_dir
        if template_default_context is not None:
            self.template_default_context = template_default_context
        else:
            self.template_default_context = {}
        self.template_default_context.setdefault('url_for', self.url_for)

        self._before_request_hooks = []
        self._before_request_hooks.extend(self.before_request_hooks)
        self._after_request_hooks = []
        self._after_request_hooks.extend(self.after_request_hooks)

        self._config = {}
        self._config.update(CONFIGURE)

    @property
    def config(self):
        return self._config

    def render_template(self, template_name, **kwargs):
        return render_template(
            template_name, template_dir=self.template_dir,
            default_context=self.template_default_context,
            context=kwargs
        ).encode('utf-8')

    def url_for(self, func_name, _request=None, _external=False, **kwargs):
        url = self._router.url_for(func_name, **kwargs)
        if _external:
            request = _request
            url = '{}://{}{}'.format(request.scheme, request.host, url)
        return url

    def url_resolve(self, path):
        """url -> view

        :return: (func, methods, func_kwargs)
        """
        return self._router.get_func(path)

    def __call__(self, environ, start_response):
        """for wsgi server"""
        self.start_response = start_response
        path = environ['PATH_INFO']
        method = environ['REQUEST_METHOD']
        func, methods, func_kwargs = self.url_resolve(path)

        try:
            if func is None:
                self.notfound()
            if method not in methods:
                self.abort(405)
            request = Request(environ)
            result = self.handle_before_request_hooks(request, view_func=func)
            if isinstance(result, Response):
                response = result
            else:
                response = self.handle_view(request, func, func_kwargs)
            self.handle_after_request_hooks(request, response, view_func=func)
        except HTTPException as ex:
            response = ex.response

        return self._start_response(response)

    def handle_view(self, request, view_func, func_kwargs):
        result = view_func(request, **func_kwargs)
        if isinstance(result, (list, tuple)):
            response = Response(content=result[0],
                                status_code=result[1],
                                headers=result[2])
        elif isinstance(result, Response):
            response = result
        else:
            response = Response(result)
        return response

    def _start_response(self, response):
        body = response.body
        status_code = response.status
        headers_list = response.headers_list
        self.start_response(status_code, headers_list)

        if isinstance(body, collections.Iterator):
            return (to_bytes(x) for x in body)
        else:
            return [to_bytes(body)]

    def route(self, path, methods=None):

        def wrapper(view_func):
            self._router.register(path, view_func, methods)
            return view_func

        return wrapper

    def add_url_rule(self, path, view_func):
        methods = view_func.methods
        self.route(path, methods=methods)(view_func)

    def before_request(self, func):
        self._before_request_hooks.append(func)
        return func

    def handle_before_request_hooks(self, request, view_func):
        hooks = self._before_request_hooks
        for hook in hooks:
            if len(inspect.signature(hook).parameters) > 1:
                result = hook(request, view_func, self)
            else:
                result = hook(request)
            if isinstance(result, Response):
                return result

    def after_request(self, func):
        self._after_request_hooks.append(func)
        return func

    def handle_after_request_hooks(self, request, response, view_func):
        hooks = self._after_request_hooks
        for hook in hooks:
            if len(inspect.signature(hook).parameters) > 2:
                hook(request, response, view_func, self)
            else:
                hook(request, response)

    def notfound(self):
        raise NotFound()

    def abort(self, code):
        raise HTTPException(Response(status_code=code))

    def make_response(self, content=b'', **kwargs):
        if isinstance(content, Response):
            return content
        return Response(content, **kwargs)

    def test_client(self):
        return Client(self)

    def run(self, host='127.0.0.1', port=5000):
        address = (host, port)
        httpd = WSGIRefServer(host, port)
        print('WSGIServer: Serving HTTP on %s ...\n' % str(address))
        httpd.run(self)


def render_template(template_name, template_dir='', default_context=None,
                    context=None, **kwargs):
    with open(os.path.join(template_dir, template_name),
              encoding='utf-8') as f:
        return Template(f.read(), context=default_context,
                        template_dir=template_dir, **kwargs
                        ).render(**context)



# -*- coding: utf-8 -*-
"""
简单的模板引擎

支持

* 直接输出变量：{{ foobar }}
* 注释: {# ... #}
* if 语句：{% if xx %} {% elif yy %} {% else %} {% endif %}
* for 循环：{% for x in lst %} {% endfor %}
* 内置函数: {{ '  foobar  '.strip() }}
* 访问对象的属性和方法: {{ foo.bar }} {{ foo.hello() }}
* 字典或列表索引: {{ foo['bar'] }}
* include: {% include "path/to/b.tpl" %}

"""
import builtins
import collections
import dis
import keyword
import os
import re

from .constants import TEMPLATE_BUILTIN_FUNC_WHITELIST
from .utils import to_text


class CodeBuilder(object):
    # 缩进步长
    INDENT_STEP = 4

    def __init__(self, indent=0):
        self.source_code = []
        # 当前缩进
        self.indent_level = indent

    def add_line(self, line):
        """增加一行代码"""
        line = ' ' * self.indent_level + line + '\n'
        self.source_code.extend([line])

    def forward_indent(self):
        """缩进前进一步"""
        self.indent_level += self.INDENT_STEP

    def back_indent(self):
        """缩进后退一步"""
        self.indent_level -= self.INDENT_STEP

    def add_section(self):
        """申请一个基于当前缩进的代码块"""
        section = CodeBuilder(self.indent_level)
        self.source_code.append(section)
        return section

    def _compile(self):
        """编译生成的代码"""
        assert self.indent_level == 0
        self._code = compile(str(self), '<source>', 'exec')
        return self._code

    def get_namespace(self):
        """执行生成的代码"""
        namespace = {}
        exec(self._code, namespace)
        return namespace

    def __str__(self):
        return ''.join(str(s) for s in self.source_code)


class Template(object):
    TOKEN_EXPR_START = '{{'
    TOKEN_EXPR_END = '}}'
    TOKEN_TAG_START = '{%'
    TOKEN_TAG_END = '%}'
    TOKEN_COMMENT_START = '{#'
    TOKEN_COMMENT_END = '#}'
    FUNC_WHITELIST = TEMPLATE_BUILTIN_FUNC_WHITELIST

    def __init__(self, text, context=None,
                 pre_compile=True,
                 indent=0, template_dir='',
                 up_vars=None,
                 func_name='render_function',
                 result_var='result',
                 auto_escape=True
                 ):
        self.tokens_re = re.compile(r'''(?sx)(
        (?:{token_expr_start}.*?{token_expr_end})
        |(?:{token_tag_start}.*?{token_tag_end})
        |(?:{token_comment_start}.*?{token_comment_end})
        )'''.format(token_expr_start=re.escape(self.TOKEN_EXPR_START),
                    token_expr_end=re.escape(self.TOKEN_EXPR_END),
                    token_tag_start=re.escape(self.TOKEN_TAG_START),
                    token_tag_end=re.escape(self.TOKEN_TAG_END),
                    token_comment_start=re.escape(self.TOKEN_COMMENT_START),
                    token_comment_end=re.escape(self.TOKEN_COMMENT_END),
                    )
        )

        self.context = {k: v
                        for k, v in builtins.__dict__.items()
                        if k in self.FUNC_WHITELIST
                        }
        self.context.update({
            'escape': escape,
            'noescape': noescape,
            'to_text': noescape,
        })
        self.base_dir = template_dir
        self.func_name = func_name
        self.result_var = result_var
        self.auto_escape = auto_escape
        # 上一层定义过的变量
        self.up_vars = up_vars or set()
        if context is not None:
            self.context.update(context)

        self.buffered = []
        self.code = code = CodeBuilder(indent=indent)
        code.add_line('def %s(context):' % func_name)
        code.forward_indent()

        # 定义 context 内的变量
        self.section_vars = code.add_section()
        # 将函数内的执行结果保存在 result 中
        code.add_line('%s = []' % result_var)
        # escape, noescape
        code.add_line('escape = context["escape"]')
        code.add_line('noescape = context["noescape"]')
        code.add_line('to_text = context["to_text"]')

        self.tpl_text = text
        # 模板中出现过的全局变量
        self.global_vars = set()
        # 模板中定义的变量
        self.tmp_vars = set()

        # 解析模板
        self.parse_text(text)

        if pre_compile:
            # 编译生成的代码
            self.code._compile()
            namespace = self.code.get_namespace()
            self.render_function = namespace[func_name]

    def parse_text(self, text):
        tokens = self.tokens_re.split(text)
        # express_stack = []

        for token in tokens:
            # 普通字符串
            if not self.tokens_re.match(token):
                self.buffered.append('%s' % repr(token))
            # {# ... #}
            elif token.startswith(self.TOKEN_COMMENT_START):
                continue
            # {{ abc }}
            elif token.startswith(self.TOKEN_EXPR_START):
                _var = self.strip_token(token, self.TOKEN_EXPR_START,
                                        self.TOKEN_EXPR_END).strip()
                self.collect_vars(_var)
                if self.auto_escape:
                    self.buffered.append('escape(%s)' % _var)
                else:
                    self.buffered.append('to_text(%s)' % _var)

            # {% blala %}
            elif token.startswith(self.TOKEN_TAG_START):
                self.flush_buffer()
                express = self.strip_token(token, self.TOKEN_TAG_START,
                                           self.TOKEN_TAG_END)
                words = express.split()
                if words[0] == 'if':   # {% if xx %}
                    _var = ' '.join(words[1:])
                    expr = 'if %s:' % _var
                    self.collect_vars(expr + 'pass')

                    self.code.add_line(expr)
                    self.code.forward_indent()
                elif words[0] == 'elif':  # {% elif xx %}
                    self.code.back_indent()
                    _var = ' '.join(words[1:])
                    expr = 'elif %s:' % _var
                    self.collect_vars(expr.replace('elif', 'if', 1) + 'pass')

                    self.code.add_line(expr)
                    self.code.forward_indent()
                elif words[0] == 'else':  # {% else %}
                    self.code.back_indent()
                    self.code.add_line('else:')
                    self.code.forward_indent()

                elif words[0] == 'for':  # {% for x in y %}
                    _var = ' '.join(words[1:])
                    expr = 'for %s :' % _var
                    self.collect_vars(expr + 'pass')
                    self.code.add_line(expr)
                    self.code.forward_indent()

                elif words[0].startswith('end'):  # {% endif %}, {% endfor %}
                    if words[0] == 'endfor':   # 排除循环过程中产生的临时变量
                        self.global_vars = self.global_vars - self.tmp_vars
                        self.tmp_vars.clear()
                    self.code.back_indent()

                elif words[0] == 'include':
                    # 保存当前 locals
                    path = ''.join(words[1:]).strip().strip('\'"')
                    func_name, _code = self.handle_include(path)
                    self.code.source_code.append(_code)
                    self.code.add_line('%s.append(%s(context))'
                                       % (self.result_var, func_name))

        self.define_global_vars()

        self.flush_buffer()
        self.code.add_line('return "".join(%s)' % self.result_var)
        self.code.back_indent()

    def define_global_vars(self):
        # 定义模板中用到的全局变量
        for name in (self.global_vars - self.tmp_vars):
            if name not in self.up_vars and not keyword.iskeyword(name):
                self.section_vars.add_line('%s = context["%s"]' % (name, name))

    def render(self, **context):
        """使用 context 字典渲染模板"""
        _context = {}
        _context.update(self.context)
        if context is not None:
            _context.update(context)

        return self.render_function(_context)

    def handle_include(self, path):
        path = os.path.join(self.base_dir, path)
        up_vars = set()
        up_vars.update(self.up_vars)
        up_vars.update(self.global_vars)
        with open(path, encoding='utf-8') as f:
            _code = Template(
                f.read(), self.context,
                pre_compile=False, indent=self.code.indent_level,
                template_dir=self.base_dir,
                up_vars=up_vars, auto_escape=self.auto_escape
            ).code
            return self.func_name, _code

    def collect_vars(self, line):
        code = compile(line, '<code>', 'exec')
        names = parse_vars(code)
        for name in names.load_names:
            self.global_vars.add(name)
        for name in names.store_names:
            self.tmp_vars.add(name)

    def flush_buffer(self):
        self.code.add_line('%s.extend([%s])'
                           % (self.result_var, ','.join(self.buffered)))
        self.buffered = []

    def strip_token(self, text, start, end):
        text = text.replace(start, '', 1)
        text = text.replace(end, '', 1)
        return text


re_parse_dis_names = re.compile('''
    (?:LOAD_NAME[^(]+
    \(
    (?P<load_name>\w+)
    \)
    )
    |
    (?:STORE_NAME[^(]+
    \(
    (?P<store_name>\w+)
    \)
    )
''', re.X)


class Writer:
    def __init__(self):
        self.data = ''

    def write(self, content):
        self.data += content


def parse_vars(code):
    writer = Writer()
    dis.dis(code, file=writer)
    names = re_parse_dis_names.findall(writer.data)
    load_names = [x[0] for x in names if x[0]]
    store_names = [x[1] for x in names if x[1]]

    return collections.namedtuple('Names', 'load_names, store_names'
                                  )(load_names, store_names)


class NoescapeText:

    def __init__(self, raw_text):
        self.raw_text = raw_text


html_escape_table = {
    '&': '&amp;',
    '"': '&quot;',
    '\'': '&apos;',
    '>': '&gt;',
    '<': '&lt;',
}


def html_escape(text):
    return ''.join(html_escape_table.get(c, c) for c in text)


def escape(text):
    if isinstance(text, NoescapeText):
        return to_text(text.raw_text)
    else:
        text = to_text(text)
        return html_escape(text)


def noescape(text):
    return NoescapeText(text)



# -*- coding: utf-8 -*-

# 模板引擎默认支持的内置函数
TEMPLATE_BUILTIN_FUNC_WHITELIST = (
    'all',
    'unicode',
    'isinstance',
    'dict',
    'format',
    'repr',
    'sorted',
    'list',
    'iter',
    'round',
    'cmp',
    'set',
    'bytes',
    'reduce',
    'slice',
    'sum',
    'getattr',
    'abs',
    'hash',
    'len',
    'ord',
    'filter',
    'range',
    'pow',
    'float',
    'divmod',
    'enumerate',
    'basestring',
    'zip',
    'hex',
    'long',
    'next',
    'chr',
    'xrange',
    'type',
    'tuple',
    'reversed',
    'hasattr',
    'delattr',
    'setattr',
    'str',
    'int',
    'unichr',
    'min',
    'any',
    'complex',
    'bool',
    'map',
    'max',
    'object',
    'callable',
)

HTTP_STATUS_CODES = {
    100:    'Continue',
    101:    'Switching Protocols',
    102:    'Processing',
    200:    'OK',
    201:    'Created',
    202:    'Accepted',
    203:    'Non Authoritative Information',
    204:    'No Content',
    205:    'Reset Content',
    206:    'Partial Content',
    207:    'Multi Status',
    226:    'IM Used',              # see RFC 3229
    300:    'Multiple Choices',
    301:    'Moved Permanently',
    302:    'Found',
    303:    'See Other',
    304:    'Not Modified',
    305:    'Use Proxy',
    307:    'Temporary Redirect',
    400:    'Bad Request',
    401:    'Unauthorized',
    402:    'Payment Required',     # unused
    403:    'Forbidden',
    404:    'Not Found',
    405:    'Method Not Allowed',
    406:    'Not Acceptable',
    407:    'Proxy Authentication Required',
    408:    'Request Timeout',
    409:    'Conflict',
    410:    'Gone',
    411:    'Length Required',
    412:    'Precondition Failed',
    413:    'Request Entity Too Large',
    414:    'Request URI Too Long',
    415:    'Unsupported Media Type',
    416:    'Requested Range Not Satisfiable',
    417:    'Expectation Failed',
    418:    'I\'m a teapot',  # see RFC 2324
    422:    'Unprocessable Entity',
    423:    'Locked',
    424:    'Failed Dependency',
    426:    'Upgrade Required',
    428:    'Precondition Required',  # see RFC 6585
    429:    'Too Many Requests',
    431:    'Request Header Fields Too Large',
    449:    'Retry With',  # proprietary MS extension
    451:    'Unavailable For Legal Reasons',
    500:    'Internal Server Error',
    501:    'Not Implemented',
    502:    'Bad Gateway',
    503:    'Service Unavailable',
    504:    'Gateway Timeout',
    505:    'HTTP Version Not Supported',
    507:    'Insufficient Storage',
    510:    'Not Extended'
}


# 默认配置
CONFIGURE = {
    'DEBUG': False,
    'SESSION_COOKIE_NAME': 'sid',
    'SESSION_COOKIE_MAX_AGE': None,
    'SESSION_COOKIE_DOMAIN': None,
    'SESSION_COOKIE_PATH': '/',
    'SESSION_COOKIE_SECURE': False,
    'SESSION_COOKIE_HTTPONLY': True,
}

NOTFOUND_HTML = b"""
<html>
    <h1>404 Not Found</h1>
</html>
"""



# -*- coding: utf-8 -*-
import collections
import re
import urllib


class Router(object):
    def __init__(self):
        self._urls_regex_map = {}
        self._urls_builer_map = {}

    def register(self, path, func, methods=None):
        url_builder = URLBuilder(path)
        url_match_re = re.compile(url_builder.url_regex)

        methods = set([x.upper() for x in methods or ['GET']])
        if 'GET' in methods and 'HEAD' not in methods:
            methods.add('HEAD')

        FuncPair = collections.namedtuple('FuncPair', ('func', 'methods'))
        self._urls_regex_map[url_match_re] = FuncPair(func, methods)
        self._urls_builer_map[url_builder] = FuncPair(func, methods)

    def get_func(self, path):
        """
        :return: (func, methods)
        """
        for url_match, func_pair in self._urls_regex_map.items():
            m = url_match.match(path)
            if m is not None:
                return func_pair.func, func_pair.methods, m.groupdict()
        return None, None, None

    def url_for(self, func_name, **kwargs):
        for url_builder, func_pair in self._urls_builer_map.items():
            func = func_pair.func
            if func.__name__ == func_name:
                return url_builder.build_url(**kwargs)
        return ''


class URLBuilder(object):
    # /<int:id>
    RE_PATH_TYPE = re.compile(r'''<
    (?:(?P<type>int|float|path):)?
    (?P<name>\w+)
    >''', re.X)
    TYPE_REGEX_MAP = {
        'int': r'\d+',
        'float': r'\d+(?:\.\d+)?',
        'path': r'.+',
    }
    # /(?P<id>\d+)
    RE_PATH_REGEX = re.compile(r'''
    \(\?P<
    (?P<name>\w+)
    >[^\)]+
    \)''', re.X)

    def __init__(self, url_exp):
        self.url_format, self.url_kwarg_names = self.exp_to_format(url_exp)
        self.url_regex = self.exp_to_regex(url_exp)

    @classmethod
    def exp_to_format(cls, exp):
        names = set()
        if cls.RE_PATH_TYPE.search(exp) or cls.RE_PATH_REGEX.search(exp):

            def replace(m):
                name = m.group('name')
                names.add(name)
                return '{' + name + '}'

            exp = cls.RE_PATH_REGEX.sub(replace, exp)
            exp = cls.RE_PATH_TYPE.sub(replace, exp)
        return exp, names

    @classmethod
    def exp_to_regex(cls, exp):
        if not exp.startswith('^'):
            if exp.startswith('/'):
                exp = '^' + exp
            else:
                exp = '^/' + exp
        if not exp.endswith('$'):
            exp = exp + '$'

        # /<int:id>
        if '(?P<' not in exp and cls.RE_PATH_TYPE.search(exp):
            exp = cls.RE_PATH_TYPE.sub(cls._replace_type_to_regex, exp)
        return exp

    @classmethod
    def _replace_type_to_regex(cls, match):
        """ /<int:id>  -> r'(?P<id>\d+)' """
        groupdict = match.groupdict()
        _type = groupdict.get('type')
        type_regex = cls.TYPE_REGEX_MAP.get(_type, '[^/]+')
        name = groupdict.get('name')
        return r'(?P<{name}>{type_regex})'.format(
            name=name, type_regex=type_regex
        )

    def build_url(self, **kwargs):
        unknown_names = set(kwargs.keys()) - self.url_kwarg_names
        url = self.url_format.format(**kwargs)
        if unknown_names:
            url += '?' + urllib.parse.urlencode(
                {k: kwargs[k] for k in unknown_names}
            )
        return url

    def __repr__(self):
        return r'<URLBuilder; url_format: {!r}>'.format(self.url_format)

    def __eq__(self, other):
        return hash(self.url_regex) == hash(other.url_regex)

    def __hash__(self):
        return hash(self.url_regex)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from codecs import open
import sys
import os

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import bustard

if sys.argv[-1] == 'publish':
    os.system('python setup.py register')
    os.system('python setup.py sdist upload')
    os.system('python setup.py bdist_wheel upload')
    sys.exit()

packages = [
    'bustard',
]

requirements = [
    # 'psycopg2',
]


def long_description():
    readme = open('README.rst', encoding='utf8').read()
    return readme

setup(
    name=bustard.__title__,
    version=bustard.__version__,
    description=bustard.__doc__,
    long_description=long_description(),
    url='https://github.com/mozillazg/bustard',
    author=bustard.__author__,
    author_email='mozillazg101@gmail.com',
    license=bustard.__license__,
    packages=packages,
    package_data={'': ['LICENSE']},
    package_dir={'bustard': 'bustard'},
    include_package_data=True,
    install_requires=requirements,
    zip_safe=False,
    platforms='any',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)
